import re
from dataclasses import dataclass
from typing import Callable, Coroutine, Optional


@dataclass
class CommandInfo:
    arguments_regex: Optional[re.Pattern]
    converters: list
    handler: Callable[..., Coroutine]



import re
from abc import ABC, abstractmethod
from typing import Any, Type


class BaseArg(ABC):

    regex: str

    def __init__(self, name):
        self.name = name

    @abstractmethod
    def get_value(self, argument_text) -> Any:
        pass


class StringArg(BaseArg):

    regex = ".+"

    def get_value(self, argument_text) -> Any:
        return argument_text


class IntArg(BaseArg):

    regex = r"\d+"

    def get_value(self, argument_text) -> Any:
        return int(argument_text)


class WordArg(BaseArg):

    regex = r"\w+"

    def get_value(self, argument_text) -> Any:
        return argument_text


class Sequence(BaseArg):

    def __init__(
            self, name, contents_type: Type[BaseArg],
            sequence_elements_separator):
        super().__init__(name)
        self.contents_type = contents_type
        self.separator = re.compile(sequence_elements_separator)
        self.regex = (
            f"{self.contents_type.regex}"
            f"(?:{sequence_elements_separator}{self.contents_type.regex})*"
        )

    def get_value(self, argument_text) -> Any:
        # noinspection PyTypeChecker
        # because None is a stub for name
        return [
            self.contents_type(None).get_value(element_str)
            for element_str in self.separator.split(argument_text)
        ]



from abc import abstractmethod, ABC
from dataclasses import dataclass
from enum import Enum, auto
from io import BytesIO
from typing import List, Iterable, Optional


class AttachmentType(Enum):
    PICTURE = auto()
    FILE = auto()
    AUDIO = auto()
    VIDEO = auto()
    OTHER = auto()


@dataclass
class Attachment:
    type: AttachmentType


class ContentCantBeDownloadedError(Exception):
    pass


class AttachmentToUploadHasUnknownFileTypeError(Exception):
    """
    Should be raised when user tries to upload an attachment with
    AttachmentType.OTHER
    """
    pass


class AbstractIncomingAttachment(Attachment, ABC):

    @abstractmethod
    async def download(self, path=None) -> Optional[bytes]:
        """
        Should return BytesIO if path is not specified,
        should raise ContentCantBeDownloadedError if attachment is not
        downloadable
        """
        pass


class UndownloadableAttachment(AbstractIncomingAttachment):

    async def download(self, path=None) -> Optional[bytes]:
        raise ContentCantBeDownloadedError()


@dataclass
class OutgoingAttachment(Attachment):
    file: BytesIO


@dataclass
class OutgoingFileAttachment(OutgoingAttachment):
    filename: str = None


@dataclass
class OutgoingMessage:
    peer_id: int
    text: Optional[str] = None
    reply_to_message_id: Optional[int] = None
    sticker_id: Optional[int] = None
    forwarded_messages_ids: Optional[Iterable[int]] = None
    attachments: Optional[List[OutgoingFileAttachment]] = None


@dataclass
class AbstractIncomingMessage(ABC):
    social_network_provider: type
    peer_id: int
    id: Optional[int] = None
    text: Optional[str] = None
    sender_id: Optional[int] = None
    sticker_id: Optional[int] = None
    attachments: Optional[List[AbstractIncomingAttachment]] = None

    @abstractmethod
    async def get_reply_message(self) -> "AbstractIncomingMessage":
        pass

    @abstractmethod
    async def get_forwarded_messages(self) -> List["AbstractIncomingMessage"]:
        pass

    def make_answer(
            self, text: Optional[str] = None,
            reply_to_message_id: Optional[int] = None,
            sticker_id: Optional[int] = None,
            forwarded_messages_ids: Optional[Iterable[int]] = None,
            attachments: Optional[List[OutgoingFileAttachment]] = None):
        return OutgoingMessage(
            peer_id=self.peer_id, text=text,
            reply_to_message_id=reply_to_message_id, sticker_id=sticker_id,
            forwarded_messages_ids=forwarded_messages_ids,
            attachments=attachments
        )

    def make_reply(
            self, text: Optional[str] = None, sticker_id: Optional[int] = None,
            forwarded_messages_ids: Optional[Iterable[int]] = None,
            attachments: Optional[List[OutgoingFileAttachment]] = None):
        return OutgoingMessage(
            peer_id=self.peer_id, text=text,
            reply_to_message_id=self.id, sticker_id=sticker_id,
            forwarded_messages_ids=forwarded_messages_ids,
            attachments=attachments
        )



def enumerate_strings(strings):
    return (
        f"{number}. {string}"
        for number, string in enumerate(strings, start=1)
    )


def enumerate_and_join_strings(strings, separator="\n"):
    return separator.join(enumerate_strings(strings))



import asyncio
import functools
import re
from typing import (
    Callable, Awaitable, Optional, List, Dict, Union, Iterable,
    Sequence
)

from mvcbotbase import helpers
from mvcbotbase.case_fold_dict import CaseFoldDict
from mvcbotbase.classes_for_command_arguments import BaseArg
from mvcbotbase.command_info import CommandInfo
from mvcbotbase.message_classes import AbstractIncomingMessage, OutgoingMessage
from mvcbotbase.social_network_provider import SocialNetworkProvider

CommandsGroupName = str
SingleCommandHelpMessage = str
CommandName = str


class MVCBotBase:
    """
    The difference between this class and others is in the language of the help
    message
    """

    def __init__(
            self, social_network_providers: (
                Union[Sequence[SocialNetworkProvider], SocialNetworkProvider]
            ),
            command_arguments_separator=" ",
            unknown_command_handler: (
                Optional[Callable[
                    [CommandName, AbstractIncomingMessage], Awaitable
                ]]
            ) = None,
            bad_command_arguments_handler: (
                Optional[Callable[
                    [CommandName, AbstractIncomingMessage], Awaitable
                ]]
            ) = None,
            handler_errors_handler: (
                Optional[Callable[[CommandInfo, asyncio.Future], Awaitable]]
            ) = None,
            help_message_title="Bot's commands"):
        if isinstance(social_network_providers, SocialNetworkProvider):
            social_network_providers = [social_network_providers]
        self.social_network_providers = social_network_providers
        self.commands_info = CaseFoldDict()
        self.command_arguments_separator = command_arguments_separator
        self.command_arguments_separator_length = len(
            command_arguments_separator
        )
        self.unknown_command_handler = unknown_command_handler
        self.bad_command_arguments_handler = bad_command_arguments_handler
        self.handler_errors_handler = handler_errors_handler
        self.command_groups: (
            Dict[CommandsGroupName, List[SingleCommandHelpMessage]]
        ) = {}
        self.ungrouped_command_help_messages: (
            Sequence[SingleCommandHelpMessage]
        ) = []
        self.help_message = None
        self.help_message_title = help_message_title

    # noinspection PyMethodMayBeStatic
    # Because maybe in future I will use `self`
    def make_help_message_for_command(
            self, names, arguments, description) -> str:
        additional_names = ", ".join([
            f"или {name}" for name in names[1:]
        ]) if len(names) > 1 else ""
        arguments = [f"[{arg.name}]" for arg in arguments]
        return " ".join(filter(None, [
            f"/{names[0]}",
            f"({additional_names})" if additional_names else None, *arguments,
            f"- {description}" if description else ""
        ]))

    def _make_full_help_message(self) -> str:
        """
        WARNING!!! Do NOT use this function to make a help message! It is used
        in library to generate the help message ONCE at launch! Generated help
        message is stored in .help_message field of your instance of MVCBotBase.
        If you wanna update help message (for example, you added/removed a
        command in runtime), call .update_message()
        """
        help_message_for_grouped_commands = "\n\n".join(
            f"• {group_name}:\n{helpers.enumerate_and_join_strings(commands)}"
            for group_name, commands in self.command_groups.items()
        )
        return "• {}:\n\n{}".format(
            self.help_message_title,
            "\n\n".join(
                [
                    help_message_for_grouped_commands,
                    helpers.enumerate_and_join_strings(
                        self.ungrouped_command_help_messages
                    )
                ]
                if self.ungrouped_command_help_messages else
                [help_message_for_grouped_commands]
            )
        )

    def add_command(
            self, name_or_names: Union[str, Sequence[str]],
            arguments: Sequence[BaseArg] = None, description=None,
            group_name=None, include_in_help_message=True, handler=None):
        """
        Works as a decorator if handler isn't specified

        group_name is used for help message (it's a command group name)
        """
        if handler:
            if isinstance(name_or_names, str):
                name_or_names = [name_or_names]
            if arguments:
                arguments_regex = re.compile(
                    self.command_arguments_separator.join(
                        f"({argument.regex})" for argument in arguments
                    )
                )
                converters = [argument.get_value for argument in arguments]
            else:
                arguments_regex = None
                converters = None
            command_info = CommandInfo(
                arguments_regex=arguments_regex, handler=handler,
                converters=converters
            )
            for name in name_or_names:
                if self.command_arguments_separator in name:
                    raise ValueError(
                        f"\"{name}\" contains arguments separator!"
                    )
                self.commands_info.setdefault(name, []).append(command_info)
            if include_in_help_message:
                help_message = self.make_help_message_for_command(
                    name_or_names, arguments if arguments else [], description
                )
                if group_name:
                    self.command_groups.setdefault(group_name, []).append(
                        help_message
                    )
                else:
                    self.ungrouped_command_help_messages.append(help_message)
            return handler
        else:
            return functools.partial(
                self.add_command, name_or_names, arguments, description,
                group_name, include_in_help_message
            )

    def remove_command(
            self, name_or_names: Union[str, Iterable[str]]):
        if isinstance(name_or_names, str):
            del self.commands_info[name_or_names]
        else:
            for name in name_or_names:
                del self.commands_info[name]

    def update_help_message(self) -> None:
        self.help_message = self._make_full_help_message()

    async def run_command_from_message(
            self, incoming_message, social_network_provider):
        command_name_length = incoming_message.text.find(
            self.command_arguments_separator
        )
        if command_name_length == -1:
            command_name = incoming_message.text
            arguments_str = ""
        else:
            command_name = incoming_message.text[:command_name_length]
            arguments_str = incoming_message.text[
                command_name_length + self.command_arguments_separator_length:
            ]
        try:
            commands_info = self.commands_info[command_name]
        except KeyError:
            if self.unknown_command_handler:
                await self.unknown_command_handler(
                    command_name, incoming_message
                )
        else:
            for command_info in commands_info:
                if command_info.arguments_regex:
                    argument_texts = command_info.arguments_regex.fullmatch(
                        arguments_str
                    )
                    if not argument_texts:
                        continue
                    answer = await command_info.handler(incoming_message, *(
                        converter(argument_text)
                        for argument_text, converter in zip(
                            argument_texts.groups(), command_info.converters
                        )
                    ))
                else:
                    if command_name_length == -1:
                        answer = await command_info.handler(incoming_message)
                    else:
                        continue
                if isinstance(answer, OutgoingMessage):
                    await social_network_provider.send_message(answer)
                elif answer is not None:
                    await social_network_provider.send_message(OutgoingMessage(
                        peer_id=incoming_message.peer_id, text=str(answer)
                    ))
                return
            if self.bad_command_arguments_handler:
                await self.bad_command_arguments_handler(
                    command_name, incoming_message
                )

    def run(self, loop: Optional[asyncio.AbstractEventLoop] = None):
        if not loop:
            loop = asyncio.get_event_loop()
        self.update_help_message()
        loop.run_until_complete(asyncio.gather(
            *[
                self._run_social_network_provider(social_network_provider)
                for social_network_provider in self.social_network_providers
            ]
        ))

    async def _run_social_network_provider(self, social_network_provider):
        async for message in social_network_provider.get_messages():
            if self.handler_errors_handler:
                asyncio.create_task(self.run_command_from_message(
                    message, social_network_provider)
                ).add_done_callback(self.handler_errors_handler)
            else:
                asyncio.create_task(
                    self.run_command_from_message(
                        message, social_network_provider
                    )
                )



from setuptools import setup

setup(
    name="mvcbotbase",
    version="0.1.1",
    packages=["mvcbotbase"],
    download_url=(
        "https://github.com/megahomyak/mvcbotbase/archive/v0.1.1.tar.gz"
    ),
    url="https://github.com/megahomyak/mvcbotbase",
    author="megahomyak",
    author_email="g.megahomyak@gmail.com",
    description="A base to make social network bots (generic)",
    install_requires=["simple_avk"]
)



from abc import ABC, abstractmethod
from typing import AsyncGenerator

from mvcbotbase.message_classes import AbstractIncomingMessage, OutgoingMessage


class SocialNetworkProvider(ABC):

    @abstractmethod
    async def get_messages(self) -> AsyncGenerator[
            AbstractIncomingMessage, None
    ]:
        yield

    @abstractmethod
    async def send_message(self, message: OutgoingMessage):
        pass



from mvcbotbase.classes_for_command_arguments import (
    BaseArg, IntArg, WordArg, StringArg, Sequence
)
from mvcbotbase.message_classes import (
    AbstractIncomingMessage, OutgoingMessage, AttachmentType,
    AbstractIncomingAttachment, OutgoingFileAttachment, Attachment,
    ContentCantBeDownloadedError, UndownloadableAttachment,
    AttachmentToUploadHasUnknownFileTypeError, OutgoingAttachment
)
from mvcbotbase.mvcbotbase import MVCBotBase
from mvcbotbase.social_network_provider import SocialNetworkProvider
